
import shm
import errno

class ShmLock (object):

    def __init__(self, keyfile, projectid=0):
        """Create a Semaphore-based Lock.

           Useful for locking resources amongst separate processes,
           not just related threads.
        
           keyfile: a (readable) file The first line should contain an
                    integer, which represents the number of instances
                    of the resource that we're locking are available.

                    This is the value the semaphore will be
                    initialized to, if semaphore creation is required.

           projectid: an integer.  It could distinguish between
                      different semaphore sets based on a single
                      keyfile.  It's probably best if you don't use
                      this.
        """

        # Create a (probably unique) SHM key by doing some hash on
        # the inode of the keyfile + projectid
        key = shm.ftok(keyfile, projectid)

        if key < 0:
            # Problem with keyfile.  Bad permissions, non-existance... something.
            # raising KeyError?  Is that a good choice?
            raise KeyError('"%s" is not a usable keyfile' % keyfile)

        if not shm.semaphore_haskey(key):
            # Semaphore doesn't exist in shared memory yet.
            # Create the thing.
            try:
                # First, get the initial value for sem from the keyfile.
                f = open(keyfile)
                val = int(f.next())
                f.close()
                del f
            except (ValueError, IOError, StopIteration), e:
                # Hmm.... this is probably dangerous. XXX
                # Do we REALLY want to assume a default for this?
                val = 1
            try:
                # Next, actually create the thing.
                shm.create_semaphore(key, val)
            except shm.error, e: # shm.error is NOT subclasses of Exception. (see NOTE)
                # This Might be dangerous to ignore.

                # I'm thinking this is some race condition where some other
                # process is trying to do the very we're doing here.

                # We can probably assume the other process was
                # successful and ignore the exception.  If it's a
                # serious problem, it will show up again later when we
                # (try to) get the semaphore.

                pass

        # Get the actual semaphore object.
        try:
            id = shm.getsemid(key)
            self._sem = shm.semaphore(id)
            self._sem.setundo(1)
        except shm.error, e:  # (see NOTE on shm.error)
            raise RuntimeError("problem getting semaphore: %s", e)

        self._acquired = 0


    def acquire(self, blocking=1):
        """Acquire resource lock.

              If blocking, return implies success, raises some exception otherwise.

              If not blocking, return True if lock is acquired, False otherwise.

              You can acquire multiple instances of a resourse, but it's not advised.
        """
        self._sem.setblocking(blocking)
        try:
            self._sem.P()
        except shm.error, e:  # (see NOTE on shm.error)
            # If not blocking, P() raises EAGAIN if resource is not available.
            if not blocking and e[0] == errno.EAGAIN:
                return False
            else:
                raise RuntimeError(str(e))
        self._acquired += 1
        return True


    def release(self):
        """Release an acquired resource lock.

           Raises a RuntimeError if an attempt is made to release a lock
           that has not been acquired.
        """
        if not self._acquired:
            raise RuntimeError("Releasing an unacquired lock.")
        self._sem.V()
        self._acquired -= 1


    def acquired(self):
        """How many instances of this resource we've acquired already."""
        return self._acquired

    
    def locked(self):
        """Test wither a lock is currently locked.  Whatever that means.

           DO NOT USE THIS.

           According to thread.LockType, 'test whether thie lock is
           currently locked' Or alternatively, according to help() on
           an actual lock, it tests 'whether the lock is in the locked
           state'.

           No idea what this means in a multiple resource locking scheme.

           I'm thinking it means 'cannot be acquire()-ed without blocking.
           That may be wrong, but it's what locked() means here.

           This is provided to conform to Python Lock types.  It probably
           shouldn't be used if the number of resource instances is not 1.
           It's all probably moot since thread.LockType isn't a proper class
           we can subtype anyway.
        """
        return self.numAvailable() <= 0


    def numWaiting(self):
        """The number of processes waiting on this resource"""
        return self._sem.ncnt


    def numAvailable(self):
        """The number of instances of this resource that are available."""
        return self._sem.val


# NOTE on shm and exceptions.
#
# When the shm module throws an exception, the type of that
# exception is shm.error, which is a string with the useless
# value "shm.error".  I tried changing shm.error to no avail.
# I tried doing 'except str, e:' also in vain.  Sadly,
# shm.error is NOT a subclass of Exception, which means that
# 'except Exception, e:' will NOT catch an exception generated
# by the shm module.
#
# The value of 'e' in 'except shm.error, e:' is (usually?)
# (always?) a tuple.
#
#    e[0] is the errno set by the library call.
#    e[1] is the perror string associated that errno.
#
# I think.

